"use strict";
/*
 * Copyright 2019 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.FirestoreBigQueryEventHistoryTracker = void 0;
const bigquery = require("@google-cloud/bigquery");
const firebase = require("firebase-admin");
const traverse = require("traverse");
const schema_1 = require("./schema");
const snapshot_1 = require("./snapshot");
const tracker_1 = require("../tracker");
const logs = require("../logs");
var schema_2 = require("./schema");
Object.defineProperty(exports, "RawChangelogSchema", { enumerable: true, get: function () { return schema_2.RawChangelogSchema; } });
Object.defineProperty(exports, "RawChangelogViewSchema", { enumerable: true, get: function () { return schema_2.RawChangelogViewSchema; } });
/**
 * An FirestoreEventHistoryTracker that exports data to BigQuery.
 *
 * When the first event is received, it creates necessary BigQuery resources:
 * - Dataset: {@link FirestoreBigQueryEventHistoryTrackerConfig#datasetId}.
 * - Table: Raw change log table {@link FirestoreBigQueryEventHistoryTracker#rawChangeLogTableName}.
 * - View: Latest view {@link FirestoreBigQueryEventHistoryTracker#rawLatestView}.
 * If any subsequent data export fails, it will attempt to reinitialize.
 */
class FirestoreBigQueryEventHistoryTracker {
    constructor(config) {
        this.config = config;
        this.initialized = false;
        this.bq = new bigquery.BigQuery();
        if (!this.config.datasetLocation) {
            this.config.datasetLocation = "us";
        }
    }
    async record(events) {
        await this.initialize();
        const rows = events.map((event) => {
            return {
                insertId: event.eventId,
                json: {
                    timestamp: event.timestamp,
                    event_id: event.eventId,
                    document_name: event.documentName,
                    document_id: event.documentId,
                    operation: tracker_1.ChangeType[event.operation],
                    data: JSON.stringify(this.serializeData(event.data)),
                },
            };
        });
        await this.insertData(rows);
    }
    serializeData(eventData) {
        if (typeof eventData === "undefined") {
            return undefined;
        }
        const data = traverse(eventData).map(function (property) {
            if (property && property.constructor) {
                if (property.constructor.name === "Buffer") {
                    this.remove();
                }
                if (property.constructor.name ===
                    firebase.firestore.DocumentReference.name) {
                    this.update(property.path);
                }
            }
        });
        return data;
    }
    /**
     * Check whether a failed operation is retryable or not.
     * Reasons for retrying:
     * 1) We added a new column to our schema. Sometimes BQ is not ready to stream insertion records immediately
     *    after adding a new column to an existing table (https://issuetracker.google.com/35905247)
     */
    async isRetryableInsertionError(e) {
        let isRetryable = true;
        const expectedErrors = [
            {
                message: "no such field.",
                location: "document_id",
            },
        ];
        if (e.response &&
            e.response.insertErrors &&
            e.response.insertErrors.errors) {
            const errors = e.response.insertErrors.errors;
            errors.forEach((error) => {
                let isExpected = false;
                expectedErrors.forEach((expectedError) => {
                    if (error.message === expectedError.message &&
                        error.location === expectedError.location) {
                        isExpected = true;
                    }
                });
                if (!isExpected) {
                    isRetryable = false;
                }
            });
        }
        return isRetryable;
    }
    /**
     * Inserts rows of data into the BigQuery raw change log table.
     */
    async insertData(rows, overrideOptions = {}, retry = true) {
        const options = {
            skipInvalidRows: false,
            ignoreUnknownValues: false,
            raw: true,
            ...overrideOptions,
        };
        try {
            const dataset = this.bigqueryDataset();
            const table = dataset.table(this.rawChangeLogTableName());
            logs.dataInserting(rows.length);
            await table.insert(rows, options);
            logs.dataInserted(rows.length);
        }
        catch (e) {
            if (retry && this.isRetryableInsertionError(e)) {
                retry = false;
                logs.dataInsertRetried(rows.length);
                return this.insertData(rows, { ...overrideOptions, ignoreUnknownValues: true }, retry);
            }
            // Reinitializing in case the destintation table is modified.
            this.initialized = false;
            throw e;
        }
    }
    /**
     * Creates the BigQuery resources with the expected schema for {@link FirestoreEventHistoryTracker}.
     * After the first invokation, it skips initialization assuming these resources are still there.
     */
    async initialize() {
        if (this.initialized) {
            return;
        }
        await this.initializeDataset();
        await this.initializeRawChangeLogTable();
        await this.initializeLatestView();
        this.initialized = true;
    }
    /**
     * Creates the specified dataset if it doesn't already exists.
     */
    async initializeDataset() {
        const dataset = this.bigqueryDataset();
        const [datasetExists] = await dataset.exists();
        if (datasetExists) {
            logs.bigQueryDatasetExists(this.config.datasetId);
        }
        else {
            logs.bigQueryDatasetCreating(this.config.datasetId);
            await dataset.create();
            logs.bigQueryDatasetCreated(this.config.datasetId);
        }
        return dataset;
    }
    /**
     * Creates the raw change log table if it doesn't already exist.
     */
    async initializeRawChangeLogTable() {
        const changelogName = this.rawChangeLogTableName();
        const dataset = this.bigqueryDataset();
        const table = dataset.table(changelogName);
        const [tableExists] = await table.exists();
        if (tableExists) {
            logs.bigQueryTableAlreadyExists(table.id, dataset.id);
            const [metadata] = await table.getMetadata();
            const fields = metadata.schema.fields;
            const documentIdColExists = fields.find((column) => column.name === "document_id");
            if (!documentIdColExists) {
                fields.push(schema_1.documentIdField);
                await table.setMetadata(metadata);
                logs.addDocumentIdColumn(this.rawChangeLogTableName());
            }
        }
        else {
            logs.bigQueryTableCreating(changelogName);
            const options = {
                friendlyName: changelogName,
                schema: schema_1.RawChangelogSchema,
            };
            if (this.config.tablePartitioning) {
                options.timePartitioning = {
                    type: this.config.tablePartitioning,
                };
            }
            await table.create(options);
            logs.bigQueryTableCreated(changelogName);
        }
        return table;
    }
    /**
     * Creates the latest snapshot view, which returns only latest operations
     * of all existing documents over the raw change log table.
     */
    async initializeLatestView() {
        const dataset = this.bigqueryDataset();
        const view = dataset.table(this.rawLatestView());
        const [viewExists] = await view.exists();
        if (viewExists) {
            logs.bigQueryViewAlreadyExists(view.id, dataset.id);
            const [metadata] = await view.getMetadata();
            const fields = metadata.schema.fields;
            const documentIdColExists = fields.find((column) => column.name === "document_id");
            if (!documentIdColExists) {
                metadata.view = snapshot_1.latestConsistentSnapshotView(this.config.datasetId, this.rawChangeLogTableName());
                await view.setMetadata(metadata);
                logs.addDocumentIdColumn(this.rawLatestView());
            }
        }
        else {
            const latestSnapshot = snapshot_1.latestConsistentSnapshotView(this.config.datasetId, this.rawChangeLogTableName());
            logs.bigQueryViewCreating(this.rawLatestView(), latestSnapshot.query);
            const options = {
                friendlyName: this.rawLatestView(),
                view: latestSnapshot,
            };
            if (this.config.tablePartitioning) {
                options.timePartitioning = {
                    type: this.config.tablePartitioning,
                };
            }
            await view.create(options);
            await view.setMetadata({ schema: schema_1.RawChangelogViewSchema });
            logs.bigQueryViewCreated(this.rawLatestView());
        }
        return view;
    }
    bigqueryDataset() {
        return this.bq.dataset(this.config.datasetId, {
            location: this.config.datasetLocation,
        });
    }
    rawChangeLogTableName() {
        return `${this.config.tableId}_raw_changelog`;
    }
    rawLatestView() {
        return `${this.config.tableId}_raw_latest`;
    }
}
exports.FirestoreBigQueryEventHistoryTracker = FirestoreBigQueryEventHistoryTracker;
